import * as createjs from "createjs-module";
import {GUI} from "dat.gui";
import * as SPLASH from "splash-screen";
import "../node_modules/splash-screen/dist/splash-screen.min.css";

import * as THREE from "three";
import {GLTFLoader} from "three/examples/jsm/loaders/GLTFLoader";
import {OrbitControls} from "three/examples/jsm/controls/OrbitControls";

import {GLTF, Utilities} from "./Three.js_utilities";

SPLASH.enable('audio-wave').then(() => {
});

let DOM_ready: ((event: Event) => void) | null = null;
Object.defineProperty(window, "DOM_ready", {
    value: new Promise(main => {
        DOM_ready = main;
    }), enumerable: false, configurable: false, writable: false
});
window.console.assert(DOM_ready !== null);
window.document.onreadystatechange = DOM_ready;

// @ts-ignore
window.DOM_ready.then(() => new Man_head());

/*
 * dat.gui
 */
class Setup {
    Color: number = 0;
}

class Man_head {
    static readonly Man_head_name = "2_Head_sculpt_retopo_mesh";
    private readonly _camera = new THREE.PerspectiveCamera(50, window.innerWidth / window.innerHeight, 0.1, 20000);
    private readonly _controls: OrbitControls;
    private readonly _renderer: THREE.WebGLRenderer = new THREE.WebGLRenderer({alpha: true});
    private readonly _scene: THREE.Scene = new THREE.Scene();

    private _tween = 0;

    constructor() {
        this._camera.position.z = 75;
        this._renderer.setClearColor(Utilities.MAGENTA, 0.5);
        this._renderer.setPixelRatio((window.devicePixelRatio) ? window.devicePixelRatio : 1);
        this._renderer.setSize(window.innerWidth, window.innerHeight);
        window.document.body.appendChild(this._renderer.domElement);
        this._scene.add(new THREE.AmbientLight(0xFFFFFF, 2.5));
        this._controls = new OrbitControls(this._camera, this._renderer.domElement);

        new Promise(ready => (new GLTFLoader()).load('./models/Man_head.gltf', this._process_GLTF.bind(this, ready))).then(value => {
            window.console.assert(value === Man_head.Man_head_name);
            this._animate();
        });
    }

    private _animate() {
        window.requestAnimationFrame(this._animate.bind(this));
        this._scene.getObjectByName("Eyes")!.rotateX(0.001);
        this._scene.getObjectByName("FRONT")!.rotateY(0.001);
        this._scene.getObjectByName("Scene")!.rotateY(0.001);
        this._renderer.render(this._scene, this._camera);
    }

    // Called within '_process_GLTF':
    private _morphing(man_head: THREE.Mesh) { // https://stackoverflow.com/questions/16361327/check-if-point-is-inside-a-custom-mesh-geometry
        const surrounding_bubble = new THREE.Group();
        this._scene.add(surrounding_bubble);
        surrounding_bubble.name = "Surrounding Bubble";
        // surrounding_bubble.visible = false;
        man_head.geometry.computeBoundingSphere();
        const sphere_geometry = new THREE.SphereGeometry(man_head.geometry.boundingSphere!.radius, 68, 40, 0, Math.PI);
        // 'raycaster' requires tiny offset so that intersections are *ALL* not empty:
        sphere_geometry.rotateY(Number.EPSILON);

        // const back = new THREE.Mesh(sphere_geometry, new THREE.MeshBasicMaterial({
        //     side: THREE.DoubleSide,
        //     wireframe: true
        // }));
        // back.rotateY(Math.PI);
        // back.name = "BACK";
        // surrounding_bubble.add(back);

        const front = new THREE.Mesh(sphere_geometry, new THREE.MeshBasicMaterial({
            alphaTest: 0.5,
            // morphNormals: true, // No longer needed with ver. 144...
            // morphTargets: true, // No longer needed with ver. 144...
            opacity: 0.5,
            side: THREE.DoubleSide,
            transparent: true
        }));
        front.name = "FRONT";
        surrounding_bubble.add(front);
        // Note that 'front' has been designed such that vertex number is close to half of 'man_head', i.e., 5784:
        window.console.assert(front.geometry.attributes.position.count === 2829);

        // Set up 'front' morphing (morphing vertex number is the same that 'front' geometry vertex number):
        const morphing = new THREE.BufferAttribute(Float32Array.from(front.geometry.attributes.position.array), front.geometry.attributes.position.itemSize);
        front.geometry.morphAttributes.position = []; // Array of each computed morphing...
        front.geometry.morphAttributes.position.push(morphing); // First one whose motion is ruled by 'front.morphTargetInfluences[0]'
        // Compute morphing (i.e., geometry of 'front' half sphere is distorted with regard to man head's face):
        const direction = new THREE.Vector3(0, 0, 0);
        const origin = new THREE.Vector3(0, 0, 0);
        const positions = front.geometry.getAttribute('position').array;

        const intersection = new Array();
        const raycaster = new THREE.Raycaster();
        for (let i = 0; i < front.geometry.attributes.position.count; i++) {
            direction.set(positions[i * 3], positions[i * 3 + 1], positions[i * 3 + 2]);
            direction.negate().normalize(); // To scene center...
            origin.set(positions[i * 3], positions[i * 3 + 1], positions[i * 3 + 2]);

            raycaster.set(origin, direction); // From point of 'front' geometry towards scene center as direction...
            /** DEBUG */
            // this._scene.add(new THREE.ArrowHelper(direction, origin));
            /** End of DEBUG */
            intersection.length = 0; // Clear...
            raycaster.intersectObject(man_head, false, intersection);
            /* Returned by 'intersectObject':
             [ { distance, point, face, faceIndex, object }, ... ]

             distance – distance between the origin of the ray and the intersection
             point – point of intersection, in world coordinates
             face – intersected face
             faceIndex – index of the intersected face
             object – the intersected object
             */
            if (intersection.length > 0) // 'intersection[0].point' is the closest point:
                // Effective distortion:
                morphing.setXYZ(i, intersection[0].point.x, intersection[0].point.y, intersection[0].point.z);
            else {
                // DEBUG: require 'sphere_geometry.rotateY(Number.EPSILON);'
                // this._scene.add(new THREE.ArrowHelper(direction, origin, 10, Utilities.RED));
                // window.console.log("origin " + JSON.stringify(origin) + " direction " + JSON.stringify(direction));
                /** End of DEBUG */
            }
        }
        front.updateMorphTargets(); // 'front.morphTargetInfluences' is reset to blank array:
        window.console.assert(front.morphTargetInfluences!.length === 1 && front.morphTargetInfluences![0] === 0);
        createjs.Tween.get(this, { // TWEENJS animation from offered functions...
            onChange: () => {
                front.morphTargetInfluences![0] = this._tween; // Morphing motion here...
            }, loop: true, paused: false
        })
            .to({_tween: 1}, 1500, createjs.Ease.linear) // Other mathematical functions are usable...
            .wait(1500)
            .to({_tween: 0}, 500, createjs.Ease.linear); // Other mathematical functions are usable...
    }

    private _process_GLTF(ready: Function, gltf: GLTF) {
        window.console.assert(gltf.scene.isGroup);
        this._scene.add(gltf.scene);
        gltf.scene.traverse((child: THREE.Object3D) => {
            if (child.type === "Mesh") {
                // window.alert(child.name + ' with type: ' + child.type);
                // "Eyes", "Plane", "Head_DL1_eyes_12_height"
                if (child.name === Man_head.Man_head_name) { // 5784 vertices
                    window.console.assert((child as THREE.Mesh).geometry.isBufferGeometry);
                    // Mandatory for morphing computation that relies on scene center:
                    (child as THREE.Mesh).geometry.center(); // Positioning...
                    // 'MeshBasicMaterial' is required to color vertices:
                    window.console.assert(((child as THREE.Mesh).material as THREE.Material).type === "MeshStandardMaterial");
                    ((child as THREE.Mesh).material as THREE.Material).dispose();
                    (child as THREE.Mesh).material = new THREE.MeshBasicMaterial({vertexColors: true, wireframe: true});
                    Utilities.Color_vertices(child as THREE.Mesh);
                    /*
                     * dat.gui
                     */
                    const dat_gui = new GUI({width: 200});
                    const max = (child as THREE.Mesh).geometry.attributes.position.count;
                    dat_gui.add(new Setup, 'Color', 0, max).step(1).onChange(Utilities.Recolor_vertices.bind(this, (child as THREE.Mesh), max));

                    this._morphing((child as THREE.Mesh));
                } else
                    child.visible = false;
                if (child.name === "Eyes_Cornea") {
                    window.console.assert(((child as THREE.Mesh).material as THREE.Material).type === "MeshStandardMaterial");
                    /** UV mapping is not set up, so don't work: */
                    // new Promise<THREE.Texture>(ready => {
                    //     new THREE.TextureLoader().load('./img/Franck.jpg', texture => {
                    //         texture.name = './img/Franck.jpg';
                    //         texture.minFilter = THREE.LinearFilter;
                    //         ready(texture);
                    //     });
                    // }).then((texture: THREE.Texture) => {
                    //     ((child as THREE.Mesh).material as THREE.MeshStandardMaterial).map = texture;
                    //     ((child as THREE.Mesh).material as THREE.MeshStandardMaterial).needsUpdate = true;
                    // });
                    //child.visible = false;
                }
                if (child.name === "Eyes")
                    child.visible = true;
            }
            if (child.type === "Scene")
                window.console.assert(child.name === child.type); // For 'getObjectByName' in animation loop...
        });
        ready(Man_head.Man_head_name);
    }
}
