import * as THREE from "three";

export interface GLTF {
    animations: Array<THREE.AnimationClip>,
    scene: THREE.Group,
    scenes: Array<THREE.Group>,
    cameras: Array<THREE.Camera>,
    asset: Object
}

export namespace Utilities {
    export const GREEN = new THREE.Color(0x00FF00);
    export const MAGENTA = new THREE.Color(0xFF00FF);
    export const RED = new THREE.Color("rgb(255,0,0)");
    export const YELLOW = new THREE.Color(0xFFFF00);

    export function Color_vertices(mesh: THREE.Mesh, color: THREE.Color = GREEN, from: number = 0, to: number = 5000) {
        (mesh.material as THREE.MeshBasicMaterial).vertexColors = true;
        (mesh.material as THREE.MeshBasicMaterial).wireframe = true;
        let colors = mesh.geometry.getAttribute('color');
        if (colors === undefined) {
            mesh.geometry.setAttribute('color', new THREE.BufferAttribute(new Float32Array(mesh.geometry.attributes.position.array.length), mesh.geometry.attributes.position.itemSize));
            colors = mesh.geometry.getAttribute('color');
        }
        // console.info(mesh.name + ": " + mesh.geometry.attributes.color.count);
        window.console.assert(to <= mesh.geometry.attributes.color.count);
        for (let i = from; i < to; i++)
            colors.setXYZ(i, color.r, color.g, color.b);
        colors.needsUpdate = true;
    }

    export function Recolor_vertices(mesh: THREE.Mesh, max: number, value: number, new_color: THREE.Color = YELLOW, old_color: THREE.Color = GREEN) {
        Color_vertices(mesh, new_color, 0, value);
        Color_vertices(mesh, old_color, value, max);
    }
}
